from .core import load_file
from .credentials import CredentialLoader
from .extract_operation import ExtractionOperationLoader
from .load_operation import LoadOperationLoader
from .state import StateLoader, save_state



import argparse
import logging
import os.path
import sys

from amaxa.loader import (
    CredentialLoader,
    ExtractionOperationLoader,
    LoadOperationLoader,
    StateLoader,
    load_file,
    save_state,
)


def main():
    a = argparse.ArgumentParser()

    a.add_argument("config", type=argparse.FileType("r"))
    a.add_argument(
        "-c",
        "--credentials",
        required=True,
        dest="credentials",
        type=argparse.FileType("r"),
    )
    a.add_argument("-l", "--load", action="store_true")
    a.add_argument("-s", "--use-state", dest="use_state", type=argparse.FileType("r"))
    a.add_argument("-k", "--check-only", dest="check_only", action="store_true")
    verbosity_levels = {
        "quiet": logging.NOTSET,
        "errors": logging.ERROR,
        "normal": logging.INFO,
        "verbose": logging.DEBUG,
    }

    a.add_argument(
        "-v",
        "--verbosity",
        choices=verbosity_levels.keys(),
        dest="verbosity",
        default="normal",
        help="Log all actions",
    )

    args = a.parse_args()

    logger = logging.getLogger("amaxa")

    logger.setLevel(verbosity_levels[args.verbosity])
    logger.handlers[:] = [logging.StreamHandler()]

    # Grab the credential file first. We need it to validate the extraction.
    credential_loader = CredentialLoader(load_file(args.credentials))
    credential_loader.load()

    if credential_loader.errors:
        errors = "\n".join(credential_loader.errors)
        logger.error(f"The supplied credentials were not valid: {errors}")
        return -1

    config = load_file(args.config)

    if args.load:
        operation_loader = LoadOperationLoader(
            config, credential_loader.result, use_state=args.use_state is not None
        )
    else:
        operation_loader = ExtractionOperationLoader(config, credential_loader.result)

    operation_loader.load()
    if operation_loader.errors:
        errors = "\n".join(operation_loader.errors)
        logger.error(f"Errors occured during load of the operation: {errors}")
        return -1

    ex = operation_loader.result

    if args.use_state:
        state_file = load_file(args.use_state)
        state_loader = StateLoader(state_file, ex)
        state_loader.load()
        if state_loader.errors:
            errors = "\n".join(state_loader.errors)
            logger.error(f"Errors occured during load of the state file: {errors}")
            return -1

    if args.check_only:
        logger.info(f"Input files validated successfully.")
        return 0

    ret = ex.run()

    if args.load and ret != 0 and ex.global_id_map:
        # Save the operation state.
        json_mode = args.config.name.endswith("json")
        with open(
            os.path.splitext(args.config.name)[0]
            + ".state"
            + (".json" if json_mode else ".yaml"),
            "w",
        ) as state_file:
            state_file.write(save_state(ex, json_mode))

    return ret


if __name__ == "__main__":
    sys.exit(main())



import csv

from .. import amaxa
from .core import OperationLoader
from .input_type import InputType


class ExtractionOperationLoader(OperationLoader):
    def __init__(self, in_dict, connection):
        super().__init__(in_dict, connection, InputType.EXTRACT_OPERATION)

    def _validate(self):
        self._validate_sobjects("queryable")
        self._validate_field_mapping()

    def _initialize(self):
        super()._initialize()
        self._open_files()

    def _load(self):
        # Create the core operation
        self.result = amaxa.ExtractOperation(self.connection)

        options = self.input.get("options") or {}

        # Create the steps and data mappers
        for entry in self.input["operation"]:
            sobject = entry["sobject"]

            mapper = self._get_data_mapper(entry, "field", "column")
            if mapper is not None:
                self.result.mappers[sobject] = mapper

            # Determine the type of extraction
            query = None
            to_extract = entry.get("extract")

            if "ids" in to_extract:
                # Register the required IDs in the context
                try:
                    for record_id in to_extract.get("ids"):
                        self.result.add_dependency(
                            sobject, amaxa.SalesforceId(record_id)
                        )
                except ValueError:
                    self.errors.append(
                        "One or more invalid Id values provided for sObject {}".format(
                            sobject
                        )
                    )

                scope = amaxa.ExtractionScope.SELECTED_RECORDS
            elif "query" in to_extract:
                query = to_extract["query"]
                scope = amaxa.ExtractionScope.QUERY
            elif "all" in to_extract:
                scope = amaxa.ExtractionScope.ALL_RECORDS
            else:
                scope = amaxa.ExtractionScope.DESCENDENTS

            field_scope = self._get_field_scope(entry)
            field_scope.add("Id")

            # Options dictionary
            step_opts = options.copy()
            step_opts.update(entry.get("options", {}))

            step = amaxa.ExtractionStep(
                sobject,
                scope,
                field_scope,
                query,
                amaxa.SelfLookupBehavior.values_dict()[entry["self-lookup-behavior"]],
                amaxa.OutsideLookupBehavior.values_dict()[
                    entry["outside-lookup-behavior"]
                ],
                options=step_opts,
            )

            self._populate_lookup_behaviors(step, entry)
            self.result.add_step(step)

    def _post_load_validate(self):
        self._validate_field_permissions()

    def _post_initialize_validate(self):
        self._validate_lookup_behaviors()

    def _get_field_scope(self, entry):
        # Use the 'field-group', 'field', and 'exclude-fields' items to derive the field scope

        fields = set()

        if "field-group" in entry:
            # Don't include types we don't process: geolocations, addresses, and base64 fields.
            if entry["field-group"] in ["readable", "smart"]:

                def include(f):
                    return f["type"] not in ["location", "address", "base64"]

            else:

                def include(f):
                    return f["createable"] and f["type"] not in [
                        "location",
                        "address",
                        "base64",
                    ]

            fields.update(
                self.result.get_filtered_field_map(entry["sobject"], include).keys()
            )

        if "fields" in entry:
            fields.update(
                {f if isinstance(f, str) else f["field"] for f in entry["fields"]}
            )

        if "exclude-fields" in entry:
            for f in entry["exclude-fields"]:
                fields.discard(f)

        return fields

    def _open_files(self):
        # Open all of the output files
        # Create DictWriters and populate them in the context
        for (step, entry) in zip(self.result.steps, self.input["operation"]):
            try:
                file_handle = open(entry["file"], "w", newline="")
                if step.sobjectname not in self.result.mappers:
                    fieldnames = step.field_scope
                else:
                    fieldnames = [
                        self.result.mappers[step.sobjectname].transform_key(k)
                        for k in step.field_scope
                    ]

                output = csv.DictWriter(
                    file_handle,
                    fieldnames=sorted(
                        fieldnames, key=lambda x: x if x != "Id" else " Id"
                    ),
                    extrasaction="ignore",
                )
                output.writeheader()
                self.result.file_store.set_file(
                    step.sobjectname, amaxa.FileType.OUTPUT, file_handle
                )
                self.result.file_store.set_csv(
                    step.sobjectname, amaxa.FileType.OUTPUT, output
                )
            except IOError as exp:
                self.errors.append(
                    "Unable to open file {} for writing ({}).".format(
                        entry["file"], exp
                    )
                )



import json

import yaml

from .. import amaxa
from .core import Loader
from .input_type import InputType


def save_state(operation, json_mode=False):
    output = {
        "version": 1,
        "state": {
            "stage": operation.stage.value,
            "id-map": {str(k): str(v) for k, v in operation.global_id_map.items()},
        },
    }

    return yaml.dump(output) if not json_mode else json.dumps(output)


class StateLoader(Loader):
    def __init__(self, in_dict, operation):
        super().__init__(in_dict, InputType.STATE)
        self.result = operation

    def _load(self):
        self.result.stage = amaxa.LoadStage.values_dict()[self.input["state"]["stage"]]
        self.result.global_id_map = {
            amaxa.SalesforceId(k): amaxa.SalesforceId(v)
            for k, v in self.input["state"]["id-map"].items()
        }



from abc import ABCMeta, abstractmethod
from typing import Dict

_all_transforms = None


class TransformException(Exception):
    pass


def get_all_transforms():
    global _all_transforms

    def get_subclasses(cls):
        classes = []
        for subclass in cls.__subclasses__():
            classes.append(subclass)
            classes.extend(get_subclasses(subclass))

        return classes

    if _all_transforms is None:
        _all_transforms = {
            cls.transform_name: cls() for cls in get_subclasses(TransformProvider)
        }

    return _all_transforms


class TransformProvider(metaclass=ABCMeta):
    transform_name = ""
    allowed_types = []

    def get_transform(self, field_context: Dict, options: Dict):
        self._validate_field(field_context)
        return self._get_transform(field_context, options)

    @abstractmethod
    def _get_transform(self, field_context: Dict, options: Dict):
        pass

    def _validate_field(self, field_context: Dict):
        if self.allowed_types and field_context["soapType"] not in self.allowed_types:
            raise TransformException(
                f"Transform {self.transform_name} is not available for fields of type {field_context['soapType']}."
            )

    def get_options_schema(self):
        return {}


class LowercaseTransformProvider(TransformProvider):
    transform_name = "lowercase"
    allowed_types = ["xsd:string"]

    def _get_transform(self, field_context: Dict, options: Dict):
        def lowercase(x):
            return x.lower()

        return lowercase


class UppercaseTransformProvider(TransformProvider):
    transform_name = "uppercase"
    allowed_types = ["xsd:string"]

    def _get_transform(self, field_context: Dict, options: Dict):
        def uppercase(x):
            return x.upper()

        return uppercase


class StripTransformProvider(TransformProvider):
    transform_name = "strip"
    allowed_types = ["xsd:string"]

    def _get_transform(self, field_context: Dict, options: Dict):
        def strip(x):
            return x.strip()

        return strip


class PrefixTransformProvider(TransformProvider):
    transform_name = "prefix"
    allowed_types = ["xsd:string"]

    def _get_transform(self, field_context: Dict, options: Dict):
        def prefix(x):
            return options["prefix"] + x

        return prefix

    def get_options_schema(self):
        return {"prefix": {"type": "string", "required": True}}


class SuffixTransformProvider(TransformProvider):
    transform_name = "suffix"
    allowed_types = ["xsd:string"]

    def _get_transform(self, field_context: Dict, options: Dict):
        def suffix(x):
            return x + options["suffix"]

        return suffix

    def get_options_schema(self):
        return {"suffix": {"type": "string", "required": True}}



import abc
import csv
import functools
import logging
from enum import Enum, unique

from . import constants


@unique
class StringEnum(Enum):
    @classmethod
    def all_values(cls):
        return [m.value for m in cls]

    @classmethod
    def values_dict(cls):
        return {m.value: m for m in cls}


class ExtractionScope(StringEnum):
    ALL_RECORDS = "all"
    QUERY = "query"
    SELECTED_RECORDS = "some"
    DESCENDENTS = "children"


class SelfLookupBehavior(StringEnum):
    TRACE_ALL = "trace-all"
    TRACE_NONE = "trace-none"


class OutsideLookupBehavior(StringEnum):
    DROP_FIELD = "drop-field"
    INCLUDE = "include"
    ERROR = "error"


class LoadStage(StringEnum):
    INSERTS = "inserts"
    DEPENDENTS = "dependents"


class FileType(Enum):
    INPUT = 1
    OUTPUT = 2
    RESULT = 3
    STATE = 4


class AmaxaException(Exception):
    pass


class SalesforceId(object):
    def __init__(self, idstr):
        if isinstance(idstr, SalesforceId):
            self.id = idstr.id
        else:
            idstr = idstr.strip()
            if len(idstr) == 15:
                suffix = ""
                for i in range(0, 3):
                    baseTwo = 0
                    for j in range(0, 5):
                        character = idstr[i * 5 + j]
                        if character >= "A" and character <= "Z":
                            baseTwo += 1 << j
                    suffix += "ABCDEFGHIJKLMNOPQRSTUVWXYZ012345"[baseTwo]
                self.id = idstr + suffix
            elif len(idstr) == 18:
                self.id = idstr
            else:
                raise ValueError("Salesforce Ids must be 15 or 18 characters.")

    def __eq__(self, other):
        if isinstance(other, SalesforceId):
            return self.id == other.id
        elif isinstance(other, str):
            return self.id == SalesforceId(other).id

        return False

    def __hash__(self):
        return hash(self.id)

    def __str__(self):
        return self.id

    def __repr__(self):
        return self.id


class FileStore(object):
    def __init__(self):
        self.store = {}
        self.csv_store = {}

    def set_file(self, sobject, ftype, f):
        self.store[(sobject, ftype)] = f

    def set_csv(self, sobject, ftype, f):
        self.csv_store[(sobject, ftype)] = f

    def get_file(self, sobject, ftype):
        return self.store[(sobject, ftype)]

    def get_csv(self, sobject, ftype):
        return self.csv_store[(sobject, ftype)]

    def close(self):
        for f in self.store.values():
            f.close()


class Operation(metaclass=abc.ABCMeta):
    def __init__(self, connection):
        self.steps = []
        self.connection = connection
        self._bulk = None
        self.logger = logging.getLogger("amaxa")
        self.file_store = FileStore()

    def run(self):
        try:
            self.initialize()
            return self.execute()
        except Exception as e:
            self.logger.error("Unexpected exception {} occurred.".format(str(e)))
            return -1
        finally:
            self.file_store.close()

    def initialize(self):
        for s in self.steps:
            s.initialize()

    @abc.abstractmethod
    def execute(self):
        pass

    def add_step(self, step):
        step.context = self
        self.steps.append(step)

    def get_sobject_list(self):
        return [step.sobjectname for step in self.steps]

    def get_sobject_name_for_id(self, id):
        return self.connection.get_sobject_name_for_id(id)

    def get_describe(self, sobjectname):
        return self.connection.get_sobject_describe(sobjectname)

    def get_field_map(self, sobjectname):
        return self.connection.get_sobject_field_map(sobjectname)

    def get_filtered_field_map(self, sobjectname, lam):
        field_map = self.get_field_map(sobjectname)

        return {k: field_map[k] for k in field_map if lam(field_map[k])}


class Step(metaclass=abc.ABCMeta):
    def __init__(self, sobjectname, field_scope):
        self.sobjectname = sobjectname
        self.field_scope = field_scope
        self.context = None
        self.options = {}

    def get_option(self, opt):
        return self.options.get(opt) or constants.OPTION_DEFAULTS[opt]

    def get_field_list(self):
        return ", ".join(self.field_scope)

    def initialize(self):
        # Determine whether we have any self-lookups or dependent lookups
        field_map = self.context.get_field_map(self.sobjectname)
        sobjects = self.context.get_sobject_list()

        # Filter for lookup fields that have at least one referent that is part of
        # this extraction. Users will see a warning for other lookups (on load),
        # and we'll just treat them like normal exported non-lookup fields
        self.all_lookups = {
            f
            for f in self.field_scope
            if field_map[f]["type"] == "reference"
            and any([s in sobjects for s in field_map[f]["referenceTo"]])
        }

        # Filter for lookup fields that are self-lookups
        # At present, we are assuming that there are no polymorphic self-lookup fields
        # in Salesforce. Should these exist, we'd have potential issues.
        self.self_lookups = {
            f
            for f in self.all_lookups
            if self.sobjectname in field_map[f]["referenceTo"]
        }

        # Filter for descendent lookups - fields that lookup to an object above this one
        # in the extraction and can be used to identify descendent records of *this* object
        self.descendent_lookups = {
            f
            for f in self.all_lookups
            if any(
                [
                    sobjects.index(refTo) < sobjects.index(self.sobjectname)
                    for refTo in field_map[f]["referenceTo"]
                    if refTo in sobjects
                ]
            )
        }

        # Filter for dependent lookups - fields that look up to an object
        # below this one in the extraction. These fields automatically have
        # dependencies registered when they're extracted.

        # A (polymorphic) field may be both a descendent lookup (up the hierarchy)
        # and a dependent lookup (down the hierarchy), as well as a lookup
        # to some arbitrary object outside the hierarchy.
        self.dependent_lookups = {
            f
            for f in self.all_lookups
            if any(
                [
                    sobjects.index(refTo) > sobjects.index(self.sobjectname)
                    for refTo in field_map[f]["referenceTo"]
                    if refTo in sobjects
                ]
            )
        }

    @abc.abstractmethod
    def execute(self):
        pass


class LoadOperation(Operation):
    def __init__(self, connection):
        super().__init__(connection)
        self.mappers = {}
        self.global_id_map = {}
        self.success = True
        self.stage = LoadStage.INSERTS

    def register_new_id(self, sobjectname, old_id, new_id):
        self.global_id_map[old_id] = new_id
        self.file_store.get_csv(sobjectname, FileType.RESULT).writerow(
            {constants.ORIGINAL_ID: str(old_id), constants.NEW_ID: str(new_id)}
        )

    def register_error(self, sobjectname, old_id, error):
        self.file_store.get_csv(sobjectname, FileType.RESULT).writerow(
            {constants.ORIGINAL_ID: str(old_id), constants.ERROR: error}
        )
        self.success = False

    def get_new_id(self, old_id):
        return self.global_id_map.get(old_id, None)

    def execute(self):
        self.logger.info(
            "Starting load with sObjects %s", ", ".join(self.get_sobject_list())
        )
        if self.stage is LoadStage.INSERTS:
            for s in self.steps:
                self.logger.info("%s: starting load", s.sobjectname)
                s.execute()

                # After each step, check whether errors happened and stop the process.
                if not self.success:
                    self.logger.error(
                        "%s: errors took place during load. See results file for details.",
                        s.sobjectname,
                    )
                    return -1

            self.stage = LoadStage.DEPENDENTS

        if self.stage is LoadStage.DEPENDENTS:
            for s in self.steps:
                self.logger.info(
                    "%s: populating dependent and self-lookups", s.sobjectname
                )
                s.execute_dependent_updates()

                if not self.success:
                    self.logger.error(
                        "%s: errors took place during dependent updates. See results file for details.",
                        s.sobjectname,
                    )
                    return -1

        return 0


class LoadStep(Step):
    def __init__(
        self,
        sobjectname,
        field_scope,
        outside_lookup_behavior=OutsideLookupBehavior.INCLUDE,
        options=None,
    ):
        self.sobjectname = sobjectname
        self.field_scope = field_scope
        self.outside_lookup_behavior = outside_lookup_behavior
        self.lookup_behaviors = {}
        self.dependent_lookup_records = []
        self.options = options or {}

        self.context = None

    def set_lookup_behavior_for_field(self, field, behavior):
        self.lookup_behaviors[field] = behavior

    def get_lookup_behavior_for_field(self, field):
        return self.lookup_behaviors.get(field, self.outside_lookup_behavior)

    def get_value_for_lookup(self, lookup, value, record_id):
        if value == "":
            return ""

        b = self.get_lookup_behavior_for_field(lookup)

        mapped_id = self.context.get_new_id(SalesforceId(value))

        if mapped_id is not None:
            return str(mapped_id)
        elif b is OutsideLookupBehavior.INCLUDE:
            return value
        elif b is OutsideLookupBehavior.ERROR:
            raise AmaxaException(
                f"{self.sobjectname} {record_id} has an outside reference in field {lookup} ({value}), "
                "which is not allowed by the extraction configuration.",
            )
        elif b is OutsideLookupBehavior.DROP_FIELD:
            return ""

    def populate_lookups(self, record, lookups, id):
        return {
            k: record[k]
            if k not in lookups
            else self.get_value_for_lookup(k, record[k], id)
            for k in record
        }

    def primitivize(self, record):
        # We're using the Bulk API over JSON, so values can be specified as strings (not converted to JSON primitives)
        # We will apply a light transformation to ensure we format correctly and respect a few Boolean equivalents
        def convert_value(value, field_type):
            if field_type == "xsd:boolean":
                if value is None or value.lower() in ["no", "false", "n", "f", "0", ""]:
                    return "false"
                elif value.lower() in ["yes", "true", "y", "t", "1"]:
                    return "true"
                raise ValueError(f"Invalid Boolean value {value}")
            elif value is None or len(value) == 0:
                return None
            elif field_type == "tns:ID":
                return str(value)
            elif field_type in [
                "xsd:string",
                "xsd:date",
                "xsd:dateTime",
                "xsd:int",
                "xsd:double",
            ]:
                return value

            return None

        field_map = self.context.get_field_map(self.sobjectname)
        return {k: convert_value(record[k], field_map[k]["soapType"]) for k in record}

    def transform_record(self, record):
        if self.sobjectname in self.context.mappers:
            record = self.context.mappers[self.sobjectname].transform_record(record)

        return {k: record[k] for k in record if k in self.field_scope}

    def clean_dependent_lookups(self, record):
        all_lookups = self.dependent_lookups | self.self_lookups

        return {k: record[k] for k in record if k not in all_lookups}

    def extract_dependent_lookups(self, record):
        all_lookups = self.dependent_lookups | self.self_lookups

        return {k: record[k] for k in record if k in all_lookups or k == "Id"}

    def execute(self):
        # Read our incoming file.
        # Apply transformations specified in our configuration file (column name -> field name, for example)
        # Then, populate all direct lookups. Dependent lookups and self-lookups will be populated in a later pass.
        records_to_load = []
        original_ids = []
        success = True

        reader = self.context.file_store.get_csv(self.sobjectname, FileType.INPUT)
        for record in reader:
            # We might have resumed this operation. Check to be sure this record hasn't been loaded already.
            if self.context.get_new_id(SalesforceId(record["Id"])) is not None:
                continue

            # We need to save off the original record Id because it'll be cleaned from the record before insert.
            # We use the original Id for error reporting.
            original_ids.append(record["Id"])

            # Then, prep this record for the Bulk API, populate its lookups, apply transforms, and clean dependent lookups
            try:
                record = self.primitivize(
                    self.populate_lookups(
                        self.clean_dependent_lookups(self.transform_record(record)),
                        self.descendent_lookups,
                        original_ids[-1],
                    )
                )
                records_to_load.append(record)
            except AmaxaException as e:
                self.context.register_error(self.sobjectname, original_ids[-1], str(e))
                success = False
            except ValueError as e:
                self.context.register_error(
                    self.sobjectname,
                    original_ids[-1],
                    f"Bad data in record {original_ids[-1]}: {str(e)}",
                )
                success = False

        if not success or len(records_to_load) == 0:
            return

        for i, r in enumerate(
            self.context.connection.bulk_api_insert(
                self.sobjectname,
                records_to_load,
                self.get_option("bulk-api-timeout"),
                self.get_option("bulk-api-poll-interval"),
                self.get_option("bulk-api-batch-size"),
            )
        ):
            if r.success:
                self.context.register_new_id(
                    self.sobjectname,
                    SalesforceId(original_ids[i]),
                    SalesforceId(r.id),  # note lowercase in result
                )
            else:
                self.context.register_error(
                    self.sobjectname, original_ids[i], self.format_error(r.error)
                )

    def execute_dependent_updates(self):
        # Populate dependent and self-lookups in a single pass
        records_to_load = []
        original_ids = []
        all_lookups = self.dependent_lookups | self.self_lookups
        success = True

        if len(all_lookups) > 0:
            # Re-check, for each record, whether we have any loading to do.
            # If all of the dependent lookups prove to be dropped outside references,
            # we have no work to do.
            self.reset_input_csv()
            reader = self.context.file_store.get_csv(self.sobjectname, FileType.INPUT)
            for record in reader:
                try:
                    cleaned_record = self.populate_lookups(
                        self.extract_dependent_lookups(record),
                        all_lookups,
                        record["Id"],
                    )
                    if (
                        len(
                            list(
                                filter(
                                    lambda r: r is not None and r != "",
                                    cleaned_record.values(),
                                )
                            )
                        )
                        > 1
                    ):  # 1 for the Id
                        # Populate the new Id for this record
                        original_ids.append(cleaned_record["Id"])
                        cleaned_record["Id"] = str(
                            self.context.get_new_id(SalesforceId(cleaned_record["Id"]))
                        )
                        records_to_load.append(cleaned_record)
                except AmaxaException as e:
                    self.context.register_error(self.sobjectname, record["Id"], str(e))
                    success = False

            if success and len(records_to_load) > 0:
                for i, r in enumerate(
                    self.context.connection.bulk_api_update(
                        self.sobjectname,
                        records_to_load,
                        self.get_option("bulk-api-timeout"),
                        self.get_option("bulk-api-poll-interval"),
                        self.get_option("bulk-api-batch-size"),
                    )
                ):
                    if not r.success:
                        self.context.register_error(
                            self.sobjectname,
                            original_ids[i],
                            self.format_error(r.error),
                        )

    def format_error(self, error):
        return "\n".join(
            [
                "{}: {}{}{}".format(
                    e["statusCode"],
                    e["message"],
                    " ({}).".format(", ".join(e["fields"]))
                    if len(e["fields"]) > 0
                    else "",
                    " " + e["extendedErrorDetails"]
                    if e["extendedErrorDetails"] is not None
                    else "",
                )
                for e in error
            ]
        )

    def reset_input_csv(self):
        fh = self.context.file_store.get_file(self.sobjectname, FileType.INPUT)
        fh.seek(0)
        self.context.file_store.set_csv(
            self.sobjectname, FileType.INPUT, csv.DictReader(fh)
        )


class ExtractOperation(Operation):
    def __init__(self, connection):
        super().__init__(connection)
        self.required_ids = {}
        self.extracted_ids = {}
        self.mappers = {}

    def execute(self):
        self.logger.info(
            "Starting extraction with sObjects %s", self.get_sobject_list()
        )
        for s in self.steps:
            self.logger.info("%s: starting extraction", s.sobjectname)
            s.execute()
            if len(s.errors) > 0:
                self.logger.error(
                    "%s: errors took place during extraction:\n%s",
                    s.sobjectname,
                    "\n".join(s.errors),
                )
                return -1
            else:
                self.logger.info(
                    "%s: extracted %d record%s",
                    s.sobjectname,
                    len(self.get_extracted_ids(s.sobjectname)),
                    "s" if len(self.get_extracted_ids(s.sobjectname)) != 1 else "",
                )

        return 0

    def add_dependency(self, sobjectname, id):
        if sobjectname not in self.required_ids:
            self.required_ids[sobjectname] = set()
        if id not in self.get_extracted_ids(sobjectname):
            self.required_ids[sobjectname].add(id)

    def get_dependencies(self, sobjectname):
        return (
            self.required_ids[sobjectname]
            if sobjectname in self.required_ids
            else set()
        )

    def get_sobject_ids_for_reference(self, sobjectname, field):
        ids = set()
        for name in self.get_field_map(sobjectname)[field]["referenceTo"]:
            # For each sObject that we've extracted data for,
            # if that object is a potential reference target for this field,
            # accumulate those Ids in a Set.
            if name in self.extracted_ids:
                ids |= self.extracted_ids[name]

        return ids

    def get_extracted_ids(self, sobjectname):
        return (
            self.extracted_ids[sobjectname]
            if sobjectname in self.extracted_ids
            else set()
        )

    def store_result(self, sobjectname, record):
        if sobjectname not in self.extracted_ids:
            self.extracted_ids[sobjectname] = set()

        if SalesforceId(record["Id"]) not in self.extracted_ids[sobjectname]:
            self.logger.debug(
                "%s: extracting record %s", sobjectname, SalesforceId(record["Id"])
            )
            self.extracted_ids[sobjectname].add(SalesforceId(record["Id"]))
            self.file_store.get_csv(sobjectname, FileType.OUTPUT).writerow(
                self.mappers[sobjectname].transform_record(record)
                if sobjectname in self.mappers
                else record
            )

        if (
            sobjectname in self.required_ids
            and SalesforceId(record["Id"]) in self.required_ids[sobjectname]
        ):
            self.required_ids[sobjectname].remove(SalesforceId(record["Id"]))


class ExtractionStep(Step):
    def __init__(
        self,
        sobjectname,
        scope,
        field_scope,
        where_clause=None,
        self_lookup_behavior=SelfLookupBehavior.TRACE_ALL,
        outside_lookup_behavior=OutsideLookupBehavior.INCLUDE,
        options=None,
    ):
        super().__init__(sobjectname, field_scope)
        self.scope = scope
        self.where_clause = where_clause
        self.self_lookup_behavior = self_lookup_behavior
        self.outside_lookup_behavior = outside_lookup_behavior
        self.lookup_behaviors = {}
        self.errors = []
        self.options = options or {}

    def set_lookup_behavior_for_field(self, f, behavior):
        self.lookup_behaviors[f] = behavior

    def get_self_lookup_behavior_for_field(self, f):
        return self.lookup_behaviors.get(f, self.self_lookup_behavior)

    def get_outside_lookup_behavior_for_field(self, f):
        return self.lookup_behaviors.get(f, self.outside_lookup_behavior)

    def execute(self):
        # If scope if ALL_RECORDS, execute a Bulk API job to extract all records
        # If scope is QUERY, execute a Bulk API job to download a query with where_clause
        # If scope is DESCENDENTS, pull based on objects that look up to any already
        # extracted object.
        # If scope is SELECTED_RECORDS, and if `context` has any registered dependencies,
        # perform a query to extract those records by Id.

        if self.scope == ExtractionScope.ALL_RECORDS:
            query = "SELECT {} FROM {}".format(self.get_field_list(), self.sobjectname)

            self.context.logger.debug(
                "%s: extracting all records using Bulk API query %s",
                self.sobjectname,
                query,
            )
            self.perform_bulk_api_pass(query)
            return
        elif self.scope == ExtractionScope.QUERY:
            query = "SELECT {} FROM {} WHERE {}".format(
                self.get_field_list(), self.sobjectname, self.where_clause
            )

            self.context.logger.debug(
                "%s: extracting filtered records using Bulk API query %s",
                self.sobjectname,
                query,
            )
            self.perform_bulk_api_pass(query)
        elif self.scope == ExtractionScope.DESCENDENTS:
            self.context.logger.debug(
                "%s: extracting descendent records based on lookups %s",
                self.sobjectname,
                ", ".join(self.descendent_lookups),
            )

            for f in self.descendent_lookups:
                self.perform_lookup_pass(f)

        # Fall through to grab all dependencies registered with the context, or SELECTED_RECORDS
        # Note that if we're tracing self-lookups, the parent objects of all extracted records so far
        # will already be registered as dependencies.

        self.resolve_registered_dependencies()

        # If we have any self-lookups, we now need to iterate to handle them.
        if (
            len(self.self_lookups) > 0
            and self.self_lookup_behavior is SelfLookupBehavior.TRACE_ALL
            and self.scope != ExtractionScope.ALL_RECORDS
        ):
            # First we query up to the parents of objects we've already obtained (i.e. the targets of their lookups)
            # Then we query down to the children of all objects obtained.
            # Then we query parents and children again.
            # We repeat until we get back no new Ids, which indicates that all references have been resolved.

            # Note that the initial parent query is handled in the dependency pass above, so we start on children.

            self.context.logger.debug(
                "%s: recursing to trace self-lookups", self.sobjectname
            )

            while True:
                before_count = len(self.context.get_extracted_ids(self.sobjectname))

                # Children
                for l in self.self_lookups:
                    self.perform_lookup_pass(l)

                # Parents
                self.resolve_registered_dependencies()

                after_count = len(self.context.get_extracted_ids(self.sobjectname))

                if before_count == after_count:
                    break

    def store_result(self, result):
        # Examine the received data to determine whether we have any cross-hierarchy lookups
        # or down-hierarchy dependencies to register

        field_map = self.context.get_field_map(self.sobjectname)
        sobject_list = self.context.get_sobject_list()

        # Add a dependency for the reference in each self lookup of this record.
        for l in self.self_lookups:
            if (
                self.get_self_lookup_behavior_for_field(l)
                is not SelfLookupBehavior.TRACE_NONE
                and result[l] is not None
            ):
                self.context.add_dependency(self.sobjectname, SalesforceId(result[l]))

        # Register any dependencies from dependent lookups
        # Note that a dependent lookup can *also* be a descendent lookup (e.g. Task.WhatId),
        # so we handle polymorphic lookups carefully
        for f in self.dependent_lookups:
            lookup_value = result[f]
            if lookup_value is not None:
                # Determine what sObject this Id looks up to
                # If this is a regular lookup, it's the target of the field, and is always dependent.
                # If this lookup is polymorphic, we have to determine it based on the Id itself,
                # and this value may actually be a cross-hierarchy reference or descendent reference.
                if len(field_map[f]["referenceTo"]) > 1:
                    target_sobject = self.context.get_sobject_name_for_id(lookup_value)

                    if target_sobject not in sobject_list:
                        continue  # Ignore references to objects not in our extraction.

                    # Determine if this is really a dependent connection, or if it's a descendent
                    # that should be handled below.
                    # The descendent code looks for cross-hierarchy references
                    if sobject_list.index(target_sobject) < sobject_list.index(
                        self.sobjectname
                    ):
                        continue

                    # Otherwise, fall through to add a dependency
                else:
                    target_sobject = field_map[f]["referenceTo"][0]

                self.context.add_dependency(target_sobject, SalesforceId(lookup_value))

        # Check for cross-hierarchy lookup values:
        # references to records above us in the extraction hierarchy, but that weren't extracted already.
        for f in self.descendent_lookups:
            lookup_value = result[f]

            if lookup_value is not None:
                if len(field_map[f]["referenceTo"]) == 1:
                    target_sobject = field_map[f]["referenceTo"][0]
                else:
                    target_sobject = self.context.get_sobject_name_for_id(lookup_value)

                if lookup_value not in self.context.get_extracted_ids(target_sobject):
                    # This is a cross-hierarchy reference
                    behavior = self.get_outside_lookup_behavior_for_field(f)

                    if behavior is OutsideLookupBehavior.DROP_FIELD:
                        del result[f]
                    elif behavior is OutsideLookupBehavior.INCLUDE:
                        continue
                    elif behavior is OutsideLookupBehavior.ERROR:
                        self.errors.append(
                            "{} {} has an outside reference in field {} ({}), which is not allowed by the extraction configuration.".format(
                                self.sobjectname, result["Id"], f, result[f]
                            )
                        )

        # Finally, call through to the context to store this result.
        self.context.store_result(self.sobjectname, result)

    def resolve_registered_dependencies(self):
        pre_deps = self.context.get_dependencies(self.sobjectname).copy()
        for r in self.context.connection.retrieve_records_by_id(
            self.sobjectname, pre_deps, self.field_scope
        ):
            self.store_result(r)

        missing = self.context.get_dependencies(self.sobjectname).intersection(pre_deps)
        if len(missing) > 0:
            self.errors.append(
                "Unable to resolve dependencies for sObject {}. The following Ids could not be found: {}".format(
                    self.sobjectname, ", ".join([str(i) for i in missing])
                )
            )

    def perform_bulk_api_pass(self, query):
        # The JSON Bulk API returns DateTime values as epoch seconds, instead of ISO 8601-format strings.
        # If we have DateTime fields in our field set, postprocess the result before we store it.
        date_time_fields = [
            f
            for f in self.field_scope
            if self.context.get_field_map(self.sobjectname)[f]["type"] == "datetime"
        ]

        for result in self.context.connection.bulk_api_query(
            self.sobjectname,
            query,
            date_time_fields,
            self.get_option("bulk-api-poll-interval"),
        ):
            self.store_result(result)

    def perform_lookup_pass(self, field):
        id_set = self.context.get_sobject_ids_for_reference(self.sobjectname, field)

        if id_set:
            for rec in self.context.connection.query_records_by_reference_field(
                self.sobjectname, self.get_field_list(), field, id_set
            ):
                self.store_result(rec)


class DataMapper(object):
    def __init__(self, field_name_mapping=None, field_transforms=None):
        self.field_name_mapping = field_name_mapping or {}
        self.field_transforms = field_transforms or {}

    def transform_record(self, record):
        return {
            self.transform_key(k): self.transform_value(k, record[k]) for k in record
        }

    def transform_key(self, k):
        return self.field_name_mapping.get(k, k)

    def transform_value(self, k, v):
        return functools.reduce(lambda x, f: f(x), self.field_transforms.get(k, []), v)



import json
import logging
import subprocess

import simple_salesforce

from .. import api, constants, jwt_auth
from .core import Loader
from .input_type import InputType


class CredentialLoader(Loader):
    def __init__(self, in_dict):
        super().__init__(in_dict, InputType.CREDENTIALS)

    def _load(self):
        if self.input["version"] == 1:
            self._load_v1()
        else:
            self._load_v2()

    def _load_v1(self):
        credentials = self.input["credentials"]

        # Determine what type of credentials we have
        if "username" in credentials and "password" in credentials:
            # User + password, optional Security Token
            self.result = simple_salesforce.Salesforce(
                username=credentials["username"],
                password=credentials["password"],
                security_token=credentials.get("security-token", ""),
                organizationId=credentials.get("organization-id", ""),
                sandbox=credentials.get("sandbox", False),
                version=constants.API_VERSION,
            )

            logging.getLogger("amaxa").debug(
                "Authenticating to Salesforce with user name and password"
            )
        elif "username" in credentials and "consumer-key" and "jwt-key" in credentials:
            # JWT authentication with key provided inline.
            try:
                self.result = jwt_auth.jwt_login(
                    credentials["consumer-key"],
                    credentials["username"],
                    credentials["jwt-key"],
                    credentials.get("sandbox", False),
                )
                logging.getLogger("amaxa").debug(
                    "Authenticating to Salesforce with inline JWT key"
                )
            except simple_salesforce.exceptions.SalesforceAuthenticationFailed as ex:
                self.errors.append(
                    "Failed to authenticate with JWT: {}".format(ex.message)
                )
        elif "username" in credentials and "jwt-file" in credentials:
            # JWT authentication with external keyfile.
            try:
                with open(credentials["jwt-file"], "r") as jwt_file:
                    self.result = jwt_auth.jwt_login(
                        credentials["consumer-key"],
                        credentials["username"],
                        jwt_file.read(),
                        credentials.get("sandbox", False),
                    )
                logging.getLogger("amaxa").debug(
                    "Authenticating to Salesforce with external JWT key"
                )
            except simple_salesforce.exceptions.SalesforceAuthenticationFailed as ex:
                self.errors.append(
                    "Failed to authenticate with JWT: {}".format(ex.message)
                )
        elif "access-token" in credentials and "instance-url" in credentials:
            self.result = simple_salesforce.Salesforce(
                instance_url=credentials["instance-url"],
                session_id=credentials["access-token"],
                version=constants.API_VERSION,
            )
            logging.getLogger("amaxa").debug(
                "Authenticating to Salesforce with access token"
            )
        else:
            self.errors.append("A set of valid credentials was not provided.")

        if self.result is not None:
            self.result = api.Connection(self.result)

    def _load_v2(self):
        credentials = self.input["credentials"]

        # Determine what type of credentials we have
        sandbox = credentials["sandbox"]
        if "username" in credentials:
            # User + password, optional Security Token
            credentials = credentials["username"]
            self.result = simple_salesforce.Salesforce(
                username=credentials["username"],
                password=credentials["password"],
                security_token=credentials.get("security-token", ""),
                organizationId=credentials.get("organization-id", ""),
                sandbox=sandbox,
                version=constants.API_VERSION,
            )

            logging.getLogger("amaxa").debug(
                "Authenticating to Salesforce with user name and password"
            )
        elif "jwt" in credentials:
            credentials = credentials["jwt"]
            if "keyfile" in credentials:
                # JWT authentication with external keyfile
                with open(credentials["keyfile"], "r") as jwt_file:
                    key = jwt_file.read()

                logging.getLogger("amaxa").debug(
                    "Authenticating to Salesforce with external JWT key"
                )
            else:
                # JWT authentication with key provided inline.
                key = credentials["key"]

            try:
                self.result = jwt_auth.jwt_login(
                    credentials["consumer-key"], credentials["username"], key, sandbox
                )
            except simple_salesforce.exceptions.SalesforceAuthenticationFailed as ex:
                self.errors.append(
                    "Failed to authenticate with JWT: {}".format(ex.message)
                )
        elif "token" in credentials:
            credentials = credentials["token"]
            self.result = simple_salesforce.Salesforce(
                instance_url=credentials["instance-url"],
                session_id=credentials["access-token"],
                version=constants.API_VERSION,
            )
            logging.getLogger("amaxa").debug(
                "Authenticating to Salesforce with access token"
            )
        elif "sfdx" in credentials:
            sfdx_username = credentials["sfdx"]

            result = subprocess.run(
                ["sfdx", "force:org:display", "--json", "-u", sfdx_username],
                encoding="utf-8",
                capture_output=True,
            )

            try:
                org_info = json.loads(result.stdout)

                if org_info["status"] == 0:
                    self.result = simple_salesforce.Salesforce(
                        instance_url=org_info["result"]["instanceUrl"],
                        session_id=org_info["result"]["accessToken"],
                        version=constants.API_VERSION,
                    )
                else:
                    self.errors.append(
                        f"SFDX failed to provide credentials with return code {org_info['status']}: {org_info['message']}."
                    )
            except Exception as e:
                self.errors.append(
                    f"SFDX failed to provide credentials with return code {result.returncode}. Exception: {e}"
                )

        if self.result is not None:
            self.result = api.Connection(self.result)

    def _post_validate(self):
        try:
            self.result.get_global_describe()
        except simple_salesforce.SalesforceError as e:
            self.errors.append("Unable to authenticate to Salesforce: {}".format(e))
            self.result = None



import importlib
import os
import sys

import cerberus

from .. import amaxa, constants, transforms
from .input_type import InputType


def get_available_versions(input_type):
    return SCHEMAS[input_type].keys()


def get_schema(input_type, version):
    return SCHEMAS[input_type][version]


def _env_or_string(params):
    ret = {
        "type": ["string", "dict"],
        "schema": {"env": {"type": "string", "required": True}},
        "coerce": lambda v: v if isinstance(v, str) else os.environ.get(v["env"]),
    }
    ret.update(params)

    return ret


def _coerce_transform(trans):
    if isinstance(trans, str):
        return {"name": trans, "options": {}}

    return trans


def _validate_import_module(field, value, error):
    try:
        if "." not in sys.path:
            sys.path.append(".")
        importlib.import_module(value)
    except ImportError:
        error(field, f"Unable to import module {value}")


def _validate_transform_options(field, value, error):
    # value will be a dict with keys "name" and "options"

    transform_name = value["name"]
    options = value["options"]

    available_transforms = transforms.get_all_transforms()

    if transform_name not in available_transforms:
        error(field, f"The transform {transform_name} does not exist.")
        return

    validator = cerberus.Validator(
        available_transforms[transform_name].get_options_schema()
    )
    validator.validate(options)

    if validator.errors:
        errors = "\n".join(validator.errors)
        error(
            field,
            f"The options schema for transform {transform_name} failed to validate: {errors}",
        )


OPTIONS_SCHEMA = {
    "type": "dict",
    "schema": {
        "bulk-api-batch-size": {
            "type": "integer",
            "default": constants.OPTION_DEFAULTS["bulk-api-batch-size"],
            "max": 10000,
            "min": 0,
        },
        "bulk-api-timeout": {
            "type": "integer",
            "default": constants.OPTION_DEFAULTS["bulk-api-timeout"],
            "min": 0,
        },
        "bulk-api-poll-interval": {
            "type": "integer",
            "default": constants.OPTION_DEFAULTS["bulk-api-poll-interval"],
            "min": 0,
            "max": 60,
        },
    },
}

SCHEMAS = {
    InputType.CREDENTIALS: {
        1: {
            "version": {"type": "integer", "required": True, "allowed": [1]},
            "credentials": {
                "type": "dict",
                "required": True,
                "schema": {
                    "username": {
                        "type": "string",
                        "excludes": ["access-token", "instance-url"],
                    },
                    "sandbox": {"type": "boolean", "default": False},
                    "access-token": {
                        "dependencies": ["instance-url"],
                        "type": "string",
                        "excludes": [
                            "username",
                            "password",
                            "security-token",
                            "jwt-key",
                            "jwt-file",
                            "consumer-key",
                        ],
                    },
                    "password": {
                        "dependencies": ["username"],
                        "type": "string",
                        "excludes": [
                            "access-token",
                            "instance-url",
                            "jwt-key",
                            "jwt-file",
                            "consumer-key",
                        ],
                    },
                    "security-token": {
                        "dependencies": ["username", "password"],
                        "type": "string",
                        "excludes": [
                            "access-token",
                            "instance-url",
                            "jwt-key",
                            "jwt-file",
                            "consumer-key",
                        ],
                    },
                    "organization-id": {
                        "dependencies": ["username", "password"],
                        "type": "string",
                        "excludes": [
                            "access-token",
                            "instance-url",
                            "jwt-key",
                            "jwt-file",
                            "consumer-key",
                        ],
                    },
                    "instance-url": {
                        "dependencies": ["access-token"],
                        "type": "string",
                        "excludes": [
                            "username",
                            "password",
                            "security-token",
                            "jwt-key",
                            "jwt-file",
                            "consumer-key",
                        ],
                    },
                    "jwt-key": {
                        "dependencies": ["consumer-key", "username"],
                        "type": "string",
                        "excludes": [
                            "password",
                            "security-token",
                            "access-token",
                            "instance-url",
                            "jwt-file",
                        ],
                    },
                    "jwt-file": {
                        "dependencies": ["consumer-key", "username"],
                        "type": "string",
                        "excludes": [
                            "password",
                            "security-token",
                            "access-token",
                            "instance-url",
                            "jwt-key",
                        ],
                    },
                    "consumer-key": {
                        "dependencies": ["username"],
                        "type": "string",
                        "excludes": [
                            "password",
                            "security-token",
                            "access-token",
                            "instance-url",
                        ],
                    },
                },
            },
        },
        2: {
            "version": {"type": "integer", "required": True, "allowed": [2]},
            "credentials": {
                "type": "dict",
                "required": True,
                "schema": {
                    "sandbox": {"type": "boolean", "default": False},
                    "username": {
                        "type": "dict",
                        "excludes": ["token", "jwt", "sfdx"],
                        "required": True,
                        "schema": {
                            "username": _env_or_string({"required": True}),
                            "password": _env_or_string({"required": True}),
                            "security-token": _env_or_string(
                                {"required": True, "excludes": "organization-id"}
                            ),
                            "organization-id": _env_or_string(
                                {"required": True, "excludes": "security-token"}
                            ),
                        },
                    },
                    "token": {
                        "type": "dict",
                        "excludes": ["username", "jwt", "sfdx"],
                        "required": True,
                        "schema": {
                            "instance-url": _env_or_string({"required": True}),
                            "access-token": _env_or_string({"required": True}),
                        },
                    },
                    "jwt": {
                        "type": "dict",
                        "excludes": ["token", "username", "sfdx"],
                        "required": True,
                        "schema": {
                            "username": _env_or_string({"required": True}),
                            "keyfile": _env_or_string(
                                {"required": True, "excludes": "key"}
                            ),
                            "key": _env_or_string(
                                {"required": True, "excludes": "keyfile"}
                            ),
                            "consumer-key": _env_or_string({"required": True}),
                        },
                    },
                    "sfdx": _env_or_string(
                        {"required": True, "excludes": ["jwt", "token", "username"],}
                    ),
                },
            },
        },
    },
    InputType.LOAD_OPERATION: {
        1: {
            "version": {"type": "integer", "required": True, "allowed": [1]},
            "operation": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": {
                        "sobject": {"type": "string", "required": True},
                        "file": {
                            "type": "string",
                            "default_setter": lambda doc: doc["sobject"] + ".csv",
                        },
                        "result-file": {
                            "type": "string",
                            "default_setter": lambda doc: doc["sobject"]
                            + "-results.csv",
                        },
                        "input-validation": {
                            "type": "string",
                            "default": "default",
                            "allowed": ["none", "default", "strict"],
                        },
                        "outside-lookup-behavior": {
                            "type": "string",
                            "allowed": amaxa.OutsideLookupBehavior.all_values(),
                            "default": "include",
                        },
                        "self-lookup-behavior": {
                            "type": "string",
                            "allowed": amaxa.SelfLookupBehavior.all_values(),
                            "default": "trace-all",
                        },
                        "extract": {
                            "type": "dict",
                            "schema": {
                                "all": {
                                    "type": "boolean",
                                    "allowed": [True],
                                    "excludes": ["descendents", "query", "ids"],
                                },
                                "descendents": {
                                    "type": "boolean",
                                    "allowed": [True],
                                    "excludes": ["all", "query", "ids"],
                                },
                                "query": {
                                    "type": "string",
                                    "excludes": ["all", "descendents", "ids"],
                                },
                                "ids": {
                                    "type": "list",
                                    "excludes": ["all", "descendents", "query"],
                                    "schema": {"type": "string"},
                                },
                            },
                        },
                        "field-group": {
                            "type": "string",
                            "allowed": ["readable", "writeable", "smart"],
                            "excludes": ["fields"],
                        },
                        "fields": {
                            "type": "list",
                            "excludes": ["field-group"],
                            "schema": {
                                "type": ["string", "dict"],
                                "schema": {
                                    "field": {"type": "string", "required": True},
                                    "column": {"type": "string", "required": False},
                                    "transforms": {
                                        "type": "list",
                                        "schema": {
                                            "type": "string",
                                            "allowed": [
                                                "strip",
                                                "lowercase",
                                                "uppercase",
                                            ],
                                            "coerce": _coerce_transform,
                                        },
                                        "required": False,
                                    },
                                    "outside-lookup-behavior": {
                                        "type": "string",
                                        "allowed": amaxa.OutsideLookupBehavior.all_values(),
                                    },
                                    "self-lookup-behavior": {
                                        "type": "string",
                                        "allowed": amaxa.SelfLookupBehavior.all_values(),
                                    },
                                },
                            },
                        },
                    },
                },
            },
        },
        2: {
            "version": {"type": "integer", "required": True, "allowed": [2]},
            "options": OPTIONS_SCHEMA,
            "plugin-modules": {
                "type": "list",
                "schema": {"type": "string", "check_with": _validate_import_module},
            },
            "operation": {
                "type": "list",
                "schema": {
                    "type": "dict",
                    "schema": {
                        "sobject": {"type": "string", "required": True},
                        "options": OPTIONS_SCHEMA,
                        "file": {
                            "type": "string",
                            "default_setter": lambda doc: doc["sobject"] + ".csv",
                        },
                        "result-file": {
                            "type": "string",
                            "default_setter": lambda doc: doc["sobject"]
                            + "-results.csv",
                        },
                        "input-validation": {
                            "type": "string",
                            "default": "default",
                            "allowed": ["none", "default", "strict"],
                        },
                        "outside-lookup-behavior": {
                            "type": "string",
                            "allowed": amaxa.OutsideLookupBehavior.all_values(),
                            "default": "include",
                        },
                        "self-lookup-behavior": {
                            "type": "string",
                            "allowed": amaxa.SelfLookupBehavior.all_values(),
                            "default": "trace-all",
                        },
                        "extract": {
                            "type": "dict",
                            "schema": {
                                "all": {
                                    "type": "boolean",
                                    "allowed": [True],
                                    "excludes": ["descendents", "query", "ids"],
                                },
                                "descendents": {
                                    "type": "boolean",
                                    "allowed": [True],
                                    "excludes": ["all", "query", "ids"],
                                },
                                "query": {
                                    "type": "string",
                                    "excludes": ["all", "descendents", "ids"],
                                },
                                "ids": {
                                    "type": "list",
                                    "excludes": ["all", "descendents", "query"],
                                    "schema": {"type": "string"},
                                },
                            },
                        },
                        "field-group": {
                            "type": "string",
                            "allowed": ["readable", "writeable", "smart"],
                        },
                        "fields": {
                            "type": "list",
                            "schema": {
                                "type": ["string", "dict"],
                                "schema": {
                                    "field": {"type": "string", "required": True},
                                    "column": {"type": "string", "required": False},
                                    "transforms": {
                                        "type": "list",
                                        "schema": {
                                            "type": ["dict", "string"],
                                            "schema": {
                                                "name": {
                                                    "type": "string",
                                                    "required": True,
                                                },
                                                "options": {"type": "dict"},
                                            },
                                            "coerce": _coerce_transform,
                                            "check_with": _validate_transform_options,
                                        },
                                        "required": False,
                                    },
                                    "outside-lookup-behavior": {
                                        "type": "string",
                                        "allowed": amaxa.OutsideLookupBehavior.all_values(),
                                    },
                                    "self-lookup-behavior": {
                                        "type": "string",
                                        "allowed": amaxa.SelfLookupBehavior.all_values(),
                                    },
                                },
                            },
                        },
                        "exclude-fields": {
                            "type": "list",
                            "schema": {"type": "string"},
                        },
                    },
                },
            },
        },
    },
    InputType.STATE: {
        1: {
            "version": {"type": "integer", "required": True, "allowed": [1]},
            "state": {
                "type": "dict",
                "required": True,
                "schema": {
                    "stage": {
                        "type": "string",
                        "required": True,
                        "allowed": amaxa.LoadStage.all_values(),
                    },
                    "id-map": {"type": "dict", "required": True},
                },
            },
        }
    },
}

SCHEMAS[InputType.EXTRACT_OPERATION] = SCHEMAS[InputType.LOAD_OPERATION]



import csv

from .. import amaxa, constants
from .core import OperationLoader
from .input_type import InputType


class LoadOperationLoader(OperationLoader):
    def __init__(self, in_dict, connection, use_state=False):
        super().__init__(in_dict, connection, InputType.LOAD_OPERATION)
        self.use_state = use_state

    def _validate(self):
        self._validate_sobjects("createable")
        self._validate_field_mapping()

    def _load(self):
        # Create the core operation
        self.result = amaxa.LoadOperation(self.connection)

        options = self.input.get("options") or {}

        # Create the steps and data mappers
        for entry in self.input["operation"]:
            sobject = entry["sobject"]

            # Mapper
            mapper = self._get_data_mapper(entry, "column", "field")
            if mapper is not None:
                self.result.mappers[sobject] = mapper
            # Field scope
            field_scope = self._get_field_scope(entry)
            # Options dictionary
            step_opts = options.copy()
            step_opts.update(entry.get("options", {}))

            step = amaxa.LoadStep(
                sobject,
                field_scope,
                amaxa.OutsideLookupBehavior.values_dict()[
                    entry["outside-lookup-behavior"]
                ],
                options=step_opts,
            )

            self._populate_lookup_behaviors(step, entry)
            self.result.add_step(step)

    def _post_load_validate(self):
        self._validate_field_permissions("createable")

    def _initialize(self):
        super()._initialize()
        self._open_files()

    def _get_field_scope(self, entry):
        # Use the 'field-group', 'field', and 'exclude-fields' items to derive the field scope

        fields = set()

        if "field-group" in entry:
            # Validation clamps acceptable values to 'writeable' or 'smart' by this point.
            # Don't include types we don't process: geolocations, addresses, and base64 fields.
            # Geolocations and addresses are omitted automatically (they aren't writeable)
            def include(f):
                return f["createable"] and f["type"] != "base64"

            fields.update(
                self.result.get_filtered_field_map(entry["sobject"], include).keys()
            )

        if "fields" in entry:
            fields.update(
                {f if isinstance(f, str) else f["field"] for f in entry["fields"]}
            )

        if "exclude-fields" in entry:
            for f in entry["exclude-fields"]:
                fields.discard(f)

        return fields

    def _open_files(self):
        # Open all of the input and output files
        # Create DictReaders and populate them in the context
        for (step, entry) in zip(self.result.steps, self.input["operation"]):
            try:
                file_handle = open(entry["file"], "r")
                input_file = csv.DictReader(file_handle)
                self.result.file_store.set_file(
                    step.sobjectname, amaxa.FileType.INPUT, file_handle
                )
                self.result.file_store.set_csv(
                    step.sobjectname, amaxa.FileType.INPUT, input_file
                )
            except IOError as exp:
                self.errors.append(
                    "Unable to open file {} for reading ({}).".format(
                        entry["file"], exp
                    )
                )

            try:
                f = open(
                    entry["result-file"], "w" if not self.use_state else "a", newline=""
                )
                output = csv.DictWriter(
                    f,
                    fieldnames=[
                        constants.ORIGINAL_ID,
                        constants.NEW_ID,
                        constants.ERROR,
                    ],
                )
                if not self.use_state:
                    output.writeheader()

                self.result.file_store.set_file(
                    step.sobjectname, amaxa.FileType.RESULT, f
                )
                self.result.file_store.set_csv(
                    step.sobjectname, amaxa.FileType.RESULT, output
                )
            except IOError as exp:
                self.errors.append(
                    "Unable to open file {} for writing ({})".format(
                        entry["result-file"], exp
                    )
                )

    def _post_initialize_validate(self):
        self._validate_dependent_field_permissions()
        self._validate_lookup_behaviors()
        self._validate_input_file_columns()

    def _validate_dependent_field_permissions(self):
        # Validate that dependent lookups are updateable.
        for step in self.result.steps:
            field_map = self.result.get_field_map(step.sobjectname)
            for f in step.dependent_lookups | step.self_lookups:
                if not field_map[f]["updateable"]:
                    self.errors.append(
                        "Field {}.{} is a dependent lookup, but is not updateable.".format(
                            step.sobjectname, f
                        )
                    )

    def _validate_input_file_columns(self):
        # Validate the column sets in the input files.
        # For each file, if validation is active, check as follows.
        # For field group steps, validate that each column in the input file
        # is mapped to a field within the field group, but allow "missing" columns.
        # For explicit field list steps, require that the mapped column set and field scope be 1:1
        for (step, entry) in zip(self.result.steps, self.input["operation"]):
            if entry["input-validation"] == "none":
                continue

            input_file = self.result.file_store.get_csv(
                step.sobjectname, amaxa.FileType.INPUT
            )
            file_field_set = set(input_file.fieldnames)
            if "Id" in file_field_set:
                file_field_set.remove("Id")

            # If we have transforms in place, transform all the column names into field names.
            if step.sobjectname in self.result.mappers:
                file_field_set = {
                    self.result.mappers[step.sobjectname].transform_key(f)
                    for f in file_field_set
                }

            if "field-group" in entry and entry["input-validation"] == "default":
                # Field group validation: file can omit columns but can't have extra
                # For the 'smart' field group, we permit any readable (but not writeable) fields
                # to be in the file, since the file was likely pulled with 'smart'=='readable'
                if entry["field-group"] == "smart":
                    comparand = set(
                        self.result.get_filtered_field_map(
                            step.sobjectname,
                            lambda f: f["type"]
                            not in ["location", "address", "base64"],
                        ).keys()
                    )
                else:
                    comparand = step.field_scope

                if not comparand.issuperset(file_field_set):
                    self.errors.append(
                        "Input file for sObject {} contains excess columns over field group '{}': {}".format(
                            step.sobjectname,
                            entry["field-group"],
                            ", ".join(sorted(file_field_set.difference(comparand))),
                        )
                    )
            else:
                # Field scope validation, or strict-mode group validation.
                # File columns must match field scope precisely.
                if step.field_scope != file_field_set:
                    self.errors.append(
                        "Input file for sObject {} does not match specified field scope.\nScope: {}\nFile Columns: {}\n".format(
                            step.sobjectname,
                            ", ".join(sorted(step.field_scope)),
                            ", ".join(sorted(file_field_set)),
                        )
                    )



import itertools
import json
from datetime import datetime, timedelta
from time import sleep
from urllib.parse import urlparse

import salesforce_bulk

from . import constants


def JSONIterator(records):
    def enc(r):
        return json.dumps(r).encode("utf-8")

    yield b"["

    i = iter(records)
    yield enc(next(i))
    for rec in i:
        yield b"," + enc(rec)

    yield b"]"


def BatchIterator(iterator, n=10000):
    while True:
        batch = list(itertools.islice(iterator, n))
        if not batch:
            return

        yield batch


class Connection(object):
    def __init__(self, sf):
        self._sf = sf
        self._bulk = salesforce_bulk.SalesforceBulk(
            sessionId=self._sf.session_id,
            host=urlparse(self._sf.bulk_url).hostname,
            API_version=constants.API_VERSION,
        )
        self._describe_info = {}
        self._field_maps = {}
        self._key_prefix_map = None

    def get_global_describe(self):
        return self._sf.describe()

    def get_sobject_field_map(self, sobjectname):
        if sobjectname not in self._describe_info:
            self.get_sobject_describe(sobjectname)

        return self._field_maps[sobjectname]

    def get_sobject_describe(self, sobjectname):
        if sobjectname not in self._describe_info:
            self._describe_info[sobjectname] = getattr(self._sf, sobjectname).describe()
            self._field_maps[sobjectname] = {
                f.get("name"): f for f in self._describe_info[sobjectname].get("fields")
            }

        return self._describe_info[sobjectname]

    def get_sobject_name_for_id(self, id):
        if self._key_prefix_map is None:
            global_describe = self.get_global_describe()["sobjects"]
            self._key_prefix_map = {
                sobject["keyPrefix"]: sobject["name"] for sobject in global_describe
            }

        return self._key_prefix_map[id[:3]]

    def _bulk_api_insert_update(
        self,
        job,
        sobject,
        record_list,
        bulk_api_timeout,
        bulk_api_poll_interval,
        bulk_api_batch_size,
    ):
        batches = []
        for record_batch in BatchIterator(iter(record_list), n=bulk_api_batch_size):
            json_iter = JSONIterator(record_batch)
            batches.append(self._bulk.post_batch(job, json_iter))

        for batch in batches:
            self._bulk.wait_for_batch(
                job,
                batch,
                timeout=bulk_api_timeout,
                sleep_interval=bulk_api_poll_interval,
            )

        self._bulk.close_job(job)

        for batch in batches:
            for r in self._bulk.get_batch_results(batch, job):
                yield r

    def bulk_api_insert(
        self,
        sobject,
        record_list,
        bulk_api_timeout,
        bulk_api_poll_interval,
        bulk_api_batch_size,
    ):
        yield from self._bulk_api_insert_update(
            self._bulk.create_insert_job(sobject, contentType="JSON"),
            sobject,
            iter(record_list),
            bulk_api_timeout,
            bulk_api_poll_interval,
            bulk_api_batch_size,
        )

    def bulk_api_update(
        self,
        sobject,
        record_list,
        bulk_api_timeout,
        bulk_api_poll_interval,
        bulk_api_batch_size,
    ):
        yield from self._bulk_api_insert_update(
            self._bulk.create_update_job(sobject, contentType="JSON"),
            sobject,
            iter(record_list),
            bulk_api_timeout,
            bulk_api_poll_interval,
            bulk_api_batch_size,
        )

    def bulk_api_query(self, sobject, query, date_time_fields, bulk_api_poll_interval):
        job = self._bulk.create_query_job(sobject, contentType="JSON")
        batch = self._bulk.query(job, query)
        self._bulk.close_job(job)

        while not self._bulk.is_batch_done(batch):
            sleep(bulk_api_poll_interval)

        for result in self._bulk.get_all_results_for_query_batch(batch):
            result = json.load(result)
            for rec in result:
                if len(date_time_fields) > 0:
                    # The JSON Bulk API returns DateTime values as epoch seconds,
                    # instead of ISO 8601-format strings.
                    # If we have DateTime fields in our field set, postprocess
                    # the result before we store it.
                    for f in date_time_fields:
                        if rec[f] is not None:
                            # Format the datetime according to Salesforce's
                            # particular wants
                            rec[f] = (
                                datetime.utcfromtimestamp(0)
                                + timedelta(milliseconds=rec[f])
                            ).isoformat(timespec="milliseconds") + "+0000"

                yield rec

    def retrieve_records_by_id(self, sobject, record_ids, field_names):
        for id_batch in BatchIterator(iter(record_ids), n=2000):
            # Make sure Ids are strings
            string_ids = [
                str(each_id) if type(each_id) is not str else each_id
                for each_id in id_batch
            ]
            for r in self._sf.restful(
                "composite/sobjects/{}".format(sobject),
                method="POST",
                data=json.dumps({"ids": string_ids, "fields": list(field_names)}),
            ):
                # None means a record with that Id is not found
                if r is not None:
                    yield r

    def query_records_by_reference_field(self, sobject, field_list, id_field, id_set):
        query = "SELECT {} FROM {} WHERE {} IN ({})"

        max_len = 4000 - len("WHERE {} IN ()".format(id_field))
        max_ids = (
            max_len // 21
        )  # 18 characters plus two quote marks and a comma, per Id

        for id_batch in BatchIterator(iter(id_set), n=max_ids):
            id_string = ",".join(["'{}'".format(str(each_id)) for each_id in id_batch])
            for r in self._sf.query_all(
                query.format(field_list, sobject, id_field, id_string)
            )["records"]:
                yield r



import datetime

import jwt
import requests
import simple_salesforce
from simple_salesforce.exceptions import SalesforceAuthenticationFailed


def jwt_login(consumer_id, username, private_key, sandbox=False):
    endpoint = (
        "https://test.salesforce.com"
        if sandbox is True
        else "https://login.salesforce.com"
    )
    jwt_payload = jwt.encode(
        {
            "exp": datetime.datetime.utcnow() + datetime.timedelta(seconds=30),
            "iss": consumer_id,
            "aud": endpoint,
            "sub": username,
        },
        private_key,
        algorithm="RS256",
    )

    result = requests.post(
        endpoint + "/services/oauth2/token",
        data={
            "grant_type": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "assertion": jwt_payload,
        },
    )
    body = result.json()

    if result.status_code != 200:
        raise SalesforceAuthenticationFailed(body["error"], body["error_description"])

    return simple_salesforce.Salesforce(
        instance_url=body["instance_url"],
        session_id=body["access_token"],
        version="46.0",
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['amaxa', 'amaxa.loader']

package_data = \
{'': ['*']}

install_requires = \
['cerberus>=1.3.2,<2.0.0',
 'cryptography>=2.8,<3.0',
 'pyjwt>=1.7.1,<2.0.0',
 'pyyaml>=5.3,<6.0',
 'requests[security]>=2.23.0,<3.0.0',
 'salesforce_bulk>=2.1.0,<3.0.0',
 'simple_salesforce>=0.75.3,<0.76.0']

entry_points = \
{'console_scripts': ['amaxa = amaxa.__main__:main']}

setup_kwargs = {
    'name': 'amaxa',
    'version': '0.9.6',
    'description': 'Load and extract data from multiple Salesforce objects in a single operation, preserving links and network structure.',
    'long_description': 'Amaxa - a multi-object ETL tool for Salesforce\n==============================================\n\n|Unit Test Badge| |Integration Test Badge| |Code Coverage Badge| |Black Badge| |PyPI Badge| |Docs Badge|\n\n.. |Unit Test Badge| image:: https://github.com/davidmreed/amaxa/workflows/Feature%20Tests/badge.svg?branch=master\n  :target: https://github.com/davidmreed/amaxa\n  :alt: Unit Tests\n\n.. |Integration Test Badge| image:: https://github.com/davidmreed/amaxa/workflows/Integration%20Test/badge.svg?branch=master\n  :target: https://github.com/davidmreed/amaxa\n  :alt: Integration Tests\n\n.. |Code Coverage Badge| image:: https://codecov.io/gh/davidmreed/amaxa/branch/master/graph/badge.svg\n  :target: https://codecov.io/gh/davidmreed/amaxa\n  :alt: Code Coverage\n\n.. |Black Badge| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n  :target: https://github.com/psf/black\n  :alt: Black Code Formatting\n\n.. |PyPI Badge| image:: https://img.shields.io/pypi/l/amaxa\n  :alt: PyPI - License\n\n.. |Docs Badge| image:: https://readthedocs.org/projects/amaxa/badge/?version=latest\n  :target: https://amaxa.readthedocs.io/en/latest/?badge=latest\n  :alt: Documentation Status\n\n\nIntroduction\n------------\n\nAmaxa is a new data loader and ETL (extract-transform-load) tool for Salesforce, designed to support the extraction and loading of complex networks of records in a single operation. For example, an Amaxa operation can extract a designated set of Accounts, their associated Contacts and Opportunities, their Opportunity Contact Roles, and associated Campaigns, and then load all of that data into another Salesforce org while preserving the connections between the records.\n\nAmaxa is designed to replace complex, error-prone workflows that manipulate data exports with spreadsheet functions like ``VLOOKUP()`` to maintain object relationships.\n\nAmaxa is free and open source software, distributed under the BSD License. Amaxa is by `David Reed <https://ktema.org>`_, (c) 2019-2020.\n\nDocumentation for Amaxa is available on `ReadTheDocs <https://amaxa.readthedocs.io>`_. The project is developed on `GitHub <https://github.com/davidmreed/amaxa>`_.\n\nWhat Does Amaxa Mean?\n---------------------\n\n`ἄμαξα <http://www.perseus.tufts.edu/hopper/text?doc=Perseus%3Atext%3A1999.04.0058%3Aentry%3Da\\)%2Fmaca>`_ is the Ancient Greek word for a wagon.\n',
    'author': 'David Reed',
    'author_email': 'david@ktema.org',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/davidmreed/amaxa',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



import abc
import collections
import json
import logging

import cerberus
import simple_salesforce
import yaml

from .. import amaxa, transforms
from . import schemas


def load_file(file_data):
    if file_data.name.endswith("json"):
        return json.load(file_data)

    return yaml.safe_load(file_data)


class Loader(metaclass=abc.ABCMeta):
    def __init__(self, in_dict, input_type):
        self.input = in_dict
        self.input_type = input_type
        self.errors = []
        self.warnings = []
        self.result = None

    def _get_validator(self):
        return cerberus.Validator(
            schemas.get_schema(self.input_type, self.input["version"])
        )

    def _validate_schema(self):
        if "version" not in self.input:
            self.errors.append("No version number present in schema")
            return

        if self.input["version"] not in schemas.get_available_versions(self.input_type):
            self.errors.append(
                "Schema version for {} not present or unsupported".format(
                    self.input_type.value
                )
            )
            return

        # Validate schema.
        validator = self._get_validator()

        self.input = validator.validated(self.input)
        if validator.errors:
            self.errors.extend(
                ["{}: {}".format(k, validator.errors[k]) for k in validator.errors]
            )

    def load(self):
        steps = [
            self._validate_schema,
            self._validate,
            self._load,
            self._post_load_validate,
            self._initialize,
            self._post_initialize_validate,
        ]

        for step in steps:
            step()
            if self.errors:
                self.result = None
                return

    def _load(self):
        pass

    def _validate(self):
        pass

    def _initialize(self):
        pass

    def _post_load_validate(self):
        pass

    def _post_initialize_validate(self):
        pass


class OperationLoader(Loader):
    def __init__(self, in_dict, connection, op_type):
        super().__init__(in_dict, op_type)
        self.connection = connection

    def _validate_field_mapping(self):
        # Validate that fields are listed only once, and that no column
        # is mapped in duplicate.
        for entry in self.input["operation"]:
            if "fields" in entry:
                field_counter = collections.Counter(
                    [f if isinstance(f, str) else f["field"] for f in entry["fields"]]
                )
                column_counter = collections.Counter(
                    [
                        f["column"] if isinstance(f, dict) and "column" in f else None
                        for f in entry["fields"]
                    ]
                )

                duplicate_fields = list(
                    filter(lambda f: field_counter[f] > 1, field_counter.keys())
                )
                duplicate_columns = list(
                    filter(
                        lambda f: f is not None and column_counter[f] > 1,
                        field_counter.keys(),
                    )
                )

                if duplicate_fields:
                    self.errors.append(
                        "{}: One or more fields is specified "
                        "multiple times: {}".format(
                            entry["sobject"], ", ".join(duplicate_fields)
                        )
                    )
                if duplicate_columns:
                    self.errors.append(
                        "{}: One or more columns is specified "
                        "multiple times: {}".format(
                            entry["sobject"], ", ".join(duplicate_columns)
                        )
                    )

    def _get_data_mapper(self, entry, source, dest):
        # Create the data mapper, if needed
        if "fields" in entry and any(
            [
                isinstance(f, dict) and ("column" in f or "transforms" in f)
                for f in entry["fields"]
            ]
        ):
            mapper = amaxa.DataMapper()

            for f in entry["fields"]:
                if isinstance(f, dict) and ("column" in f or "transforms" in f):
                    if "column" in f:
                        mapper.field_name_mapping[f[source]] = f[dest]

                    if "transforms" in f:
                        try:
                            field_name = f[source] if source in f else f[dest]
                            mapper.field_transforms[field_name] = [
                                self._build_transform(entry["sobject"], f["field"], t)
                                for t in f["transforms"]
                            ]
                        except transforms.TransformException as e:
                            self.errors.append(
                                f"Unable to create transforms for field {entry['sobject']}.{f['field']}: {e}"
                            )

            return mapper

        return None

    def _build_transform(self, sobject_name, field_name, transform):
        transform_factory = transforms.get_all_transforms()[transform["name"]]

        return transform_factory.get_transform(
            self.connection.get_sobject_field_map(sobject_name)[field_name],
            transform["options"],
        )

    def _populate_lookup_behaviors(self, step, entry):
        if "fields" in entry and any(
            [
                isinstance(f, dict)
                and ("self-lookup-behavior" in f or "outside-lookup-behavior" in f)
                for f in entry["fields"]
            ]
        ):
            for f in entry["fields"]:
                if isinstance(f, dict) and (
                    "self-lookup-behavior" in f or "outside-lookup-behavior" in f
                ):
                    if "self-lookup-behavior" in f:
                        step.set_lookup_behavior_for_field(
                            f["field"],
                            amaxa.SelfLookupBehavior.values_dict()[
                                f["self-lookup-behavior"]
                            ],
                        )
                    if "outside-lookup-behavior" in f:
                        step.set_lookup_behavior_for_field(
                            f["field"],
                            amaxa.OutsideLookupBehavior.values_dict()[
                                f["outside-lookup-behavior"]
                            ],
                        )

    def _validate_field_permissions(self, permission=None):
        # Validate that all fields designated for steps are real, writeable,
        # and of a supported type.
        all_sobjects = [step.sobjectname for step in self.result.steps]

        for step in self.result.steps:
            field_map = self.result.get_field_map(step.sobjectname)
            for field in step.field_scope:
                if field not in field_map or (
                    permission is not None and not field_map[field][permission]
                ):
                    self.errors.append(
                        "Field {}.{} does not exist or does not "
                        "have the correct CRUD permission{}.".format(
                            step.sobjectname,
                            field,
                            " (" + permission + ")" if permission else "",
                        )
                    )
                elif field_map[field]["type"] == "reference":
                    # Ensure that the target objects of this reference
                    # are included in the operation. If not, show a warning.
                    if not any(
                        [ref in all_sobjects for ref in field_map[field]["referenceTo"]]
                    ):
                        logging.getLogger("amaxa").warning(
                            "Field %s.%s is a reference none of whose targets (%s) "
                            "are included in the operation. Reference handlers "
                            "will be inactive for references to non-included sObjects.",
                            step.sobjectname,
                            field,
                            ", ".join(field_map[field]["referenceTo"]),
                        )
                    elif not all(
                        [ref in all_sobjects for ref in field_map[field]["referenceTo"]]
                    ):
                        logging.getLogger("amaxa").debug(
                            "Field %s.%s is a reference whose targets (%s) "
                            "are not all included in the operation. Reference handlers "
                            "will be inactive for references to non-included sObjects.",
                            step.sobjectname,
                            field,
                            ", ".join(field_map[field]["referenceTo"]),
                        )
                elif field_map[field]["type"] in ["location", "address", "base64"]:
                    self.errors.append(
                        "Field {}.{} is of an unsupported type ({})".format(
                            step.sobjectname, field, field_map[field]["type"]
                        )
                    )

    def _validate_sobjects(self, permission):
        try:
            global_describe = {
                entry["name"]: entry
                for entry in self.connection.get_global_describe()["sobjects"]
            }
        except simple_salesforce.SalesforceAuthenticationFailed as e:
            self.errors.append("Unable to authenticate to Salesforce: {}".format(e))
            return

        for entry in self.input["operation"]:
            sobject = entry["sobject"]

            if (
                sobject not in global_describe
                or not global_describe[sobject][permission]
            ):
                self.errors.append(
                    "sObject {} does not exist or does not "
                    "have the correct permission ({})".format(sobject, permission)
                )

    def _validate_lookup_behaviors(self):
        # Validate that lookup behaviors are associated with lookups
        # of the correct type (outside or self)
        for step in self.result.steps:
            for f in step.lookup_behaviors:
                if (
                    f in step.dependent_lookups
                    and step.lookup_behaviors[f] not in amaxa.OutsideLookupBehavior
                ) or (
                    f in step.self_lookups
                    and step.lookup_behaviors[f] not in amaxa.SelfLookupBehavior
                ):
                    self.errors.append(
                        "Lookup behavior '{}' specified for field {}."
                        "{} is not valid for this lookup type.".format(
                            step.lookup_behaviors[f].value, step.sobjectname, f
                        )
                    )

    def _initialize(self):
        self.result.initialize()



ORIGINAL_ID = "Original Id"
NEW_ID = "New Id"
ERROR = "Error"
OPTION_DEFAULTS = {
    "bulk-api-poll-interval": 5,
    "bulk-api-timeout": 1200,
    "bulk-api-batch-size": 10000,
}
API_VERSION = "46.0"
